# -*- coding: utf-8 -*-
# verbinder.py


import platform
versionsNummer=platform.python_version()[0]
import math

from klassendiagramme.figuren import *
from klassendiagramme.klassenkarte import Klassenkarte

### ----- globale Hilfsfunktion -----------------------
def Zwischen(x, klein, gross): return x>=klein and x<=gross

### ---------------------------------------------------
class Verbinder(object):
    """Linie oder Pfeil fuer UML-Diagramme
    Verbinder wird wie eine abstrakte Klasse behandelt."""

    def __init__(self,
                 x0=100, y0=100, x1=200, y1=150,
                 name='',
                 richtung=0,
                 typ=0):
        self.__x0,self.__y0,self.__x1,self.__y1 = x0,y0,x1,y1
        self.__name = name
        self.__richtung=richtung
        self.__typ=typ
        self.SetzeFigur(self.GibFigur())
        self.Zeige()

    ## Get-Methoden
    def GibStart(self): return self.__x0,self.__y0
    def GibZiel(self): return self.__x1,self.__y1
    def GibDaten(self): return self.__x0,self.__y0,self.__x1,self.__y1
    def GibNamen(self): return self.__name
    def GibRichtung(self): return self.__richtung
    def GibTyp(self): return self.__typ

    def GibFigur(self):
        '''definiert die Figur als waagerechten Pfeil, aber mit
        richtiger Laenge und setzt den Winkel richtig'''
        x0,y0,x1,y1=self.GibDaten()
        laenge=math.sqrt((x1-x0)**2+(y1-y0)**2)
        if laenge<30: laenge=30 # Pfeil zu kurz
        if (int(x1-x0)==0) and (y1>y0): winkel=90
        elif (int(x1-x0)==0) and (y1<=y0): winkel=270
        elif x1>x0:
            winkel=math.atan((y1-y0)/float(x1-x0))*180/math.pi
        else:
            winkel=180+math.atan((y1-y0)/float(x1-x0))*180/math.pi
        figur=FigurenGruppe2(x0,y0,winkel)
        figur.FuegeHinzu(Linie(0, 0, laenge, 0)) # Linie braucht breite und hoehe
        if self.__richtung==1:
            figur.FuegeHinzu(
                N_Eck(laenge-15,0,((0,-7),(15,0),(0,7)),
                      geschlossen=False))
        if self.__typ==1:
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(15,-5),(30,0),(15,5)),fuellfarbe='white'))
        elif self.__typ==2:
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(15,-5),(30,0),(15,5)),fuellfarbe='black'))
        if self.__typ==3:
            figur.FuegeHinzu(
                N_Eck(laenge,0,((-20,-8),(0,0),(-20,8)),fuellfarbe='white'))
        return figur

    def SetzeStart(self, koordinaten):
        self.__x0,self.__y0=koordinaten
        self.__figur.Verberge()
        self.SetzeFigur(self.GibFigur())
        self.__figur.Zeige()

    def SetzeZiel(self, koordinaten):
        self.__x1,self.__y1=koordinaten
        self.__figur.Verberge()
        self.SetzeFigur(self.GibFigur())
        self.__figur.Zeige()

    def SetzeFigur(self, figur):
        self.__figur=figur

    def SetzeRichtung(self, richtung):
        '''Setzt Pfeil oder nicht neu'''
        self.Verberge()
        self.__richtung=richtung
        self.SetzeFigur(self.GibFigur())
        self.Zeige()

    def SetzeTyp(self, typ):
        '''setzt den Typ neu'''
        if typ in (0,1,2,3):
            self.Verberge()
            self.__typ=typ
            self.SetzeFigur(self.GibFigur())
            self.Zeige()

    def Verberge(self):
        if self.__figur==None: return
        self.__figur.Verberge()
        
    def Zeige(self):
        if self.__figur==None:
            self.__figur=self.GibFigur()
        self.__figur.Zeige()


### ----- verbinder1 ---------------------------------------
## rechtwinklige Verbindung mit einem Winkel
class Verbinder1(Verbinder):
    """Linie oder Pfeil fuer UML-Diagramme
    In dieser Variante werden zwei zu einander senkrechte Linien
    verwendet"""

    def __init__(self,
                 x0, y0, x1, y1,
                 name='',
                 anschluss='NO', # alle Kombinationen moeglich
                 richtung=0, # Pfeil?
                 typ=0
                 ): 
        self.__anschluss=anschluss # wie in die Karte hinein,
        ## also die Orientierung der zweiten Linie
        Verbinder.__init__(self,x0,y0,x1,y1,name,richtung,typ)
        # Figur als FigurenGruppe initialisieren

    def GibFigur(self):
        '''setzt die aktuelle Figur'''
        x0,y0,x1,y1 = self.GibDaten()
        b,h=x1-x0,y1-y0
        figur=FigurenGruppe(x0,y0)
        if self.__anschluss[1]=='W': # von links hinein
            figur.FuegeHinzu(Linie(0,0,0,h))
            figur.FuegeHinzu(Linie(0,h,b,0))
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-15,-7),(0,0),(-15,7)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-20,-8),(0,0),(-20,8)),
                          fuellfarbe='white'))
        elif self.__anschluss[1]=='O': # von rechts hinein
            figur.FuegeHinzu(Linie(0,0,0,h))
            figur.FuegeHinzu(Linie(0,h,b,0))
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((15,-7),(0,0),(15,7)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((20,-8),(0,0),(20,8)),
                          fuellfarbe='white'))
        elif self.__anschluss[1]=='N': # von oben hinein
            figur.FuegeHinzu(Linie(0,0,b,0))
            figur.FuegeHinzu(Linie(b,0,0,h))
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-7,-15),(0,0),(7,-15)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-8,-20),(0,0),(8,-20)),
                          fuellfarbe='white'))
        else:                   # 'S'
            figur.FuegeHinzu(Linie(0,0,b,0))
            figur.FuegeHinzu(Linie(b,0,0,h))
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-7,15),(0,0),(7,15)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-8,20),(0,0),(8,20)),
                          fuellfarbe='white'))
        # Raute ?
        if not(self.GibTyp() in (1,2)): return figur
        if self.GibTyp()==1: ff='white'
        if self.GibTyp()==2: ff='black'
        if self.__anschluss[0]=='O':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(15,-5),(30,0),(15,5)),fuellfarbe=ff))
        elif self.__anschluss[0]=='W':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-15,-5),(-30,0),(-15,5)),fuellfarbe=ff))
        elif self.__anschluss[0]=='S':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-5,15),(0,30),(5,15)),fuellfarbe=ff))
        else:                   # 'N'
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-5,-15),(0,-30),(5,-15)),fuellfarbe=ff))
        return figur


### ----- verbinder2 ---------------------------------------
## rechtwinklige Verbindung mit zwei Winkeln
class Verbinder2(Verbinder):
    """Linie oder Pfeil fuer UML-Diagramme
    In dieser Variante werden drei zu einander senkrechte Linien
    verwendet
    Voraussetzung ist O->W, N->S oder umgekehrt"""

    def __init__(self,
                 x0, y0, x1, y1,
                 name='',
                 anschluss='OW',  # alternativ 'WO', 'NS' oder 'SN'
                 richtung=0, # Pfeil?
                 typ=0
                 ):
        self.__anschluss=anschluss
        ## die anderen Informationen ergeben sich aus den Koordinaten
        self.__teiler=0.5 # standardmaessig die Haelfte
        Verbinder.__init__(self,x0,y0,x1,y1,name,richtung,typ)
        # Figur als FigurenGruppe initialisieren

    def AendereTeiler(self,teiler):
        self.Verberge()
        self.__teiler=teiler
        self.Zeige()

    def GibFigur(self):
        '''gibt die aktuelle Figur zurueck'''
        x0,y0,x1,y1 = self.GibDaten()
        b,h=x1-x0,y1-y0
        figur=FigurenGruppe(x0,y0)
        if self.__anschluss in ('OW','WO'):
            xz_relativ=int((x1-x0)*self.__teiler) 
            figur.FuegeHinzu(Linie(0,0,xz_relativ,0))
            figur.FuegeHinzu(Linie(xz_relativ,0,0,h))
            figur.FuegeHinzu(Linie(xz_relativ,h,b-xz_relativ,0))
        else:   # 'NS' oder 'SN'
            yz_relativ=int((y1-y0)*self.__teiler)
            figur.FuegeHinzu(Linie(0,0,0,yz_relativ))
            figur.FuegeHinzu(Linie(0,yz_relativ,b,0))
            figur.FuegeHinzu(Linie(b,yz_relativ,0,h-yz_relativ))
        if self.__anschluss[1]=='N':
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-7,-15),(0,0),(7,-15)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-8,-20),(0,0),(8,-20)),
                          fuellfarbe='white'))
        elif self.__anschluss[1]=='S':
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-7,15),(0,0),(7,15)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-8,20),(0,0),(8,20)),
                          fuellfarbe='white'))
        elif self.__anschluss[1]=='W':
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((-15,-7),(0,0),(-15,7)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((-20,-8),(0,0),(-20,8)),
                          fuellfarbe='white'))
        else:                   # 'O'
            if self.GibRichtung()==1: # Pfeil zeichnen
                figur.FuegeHinzu(
                    N_Eck(b,h,((15,-7),(0,0),(15,7)),
                          geschlossen=False))
            elif self.GibTyp()==3: # Vererbungspfeil
                figur.FuegeHinzu(
                    N_Eck(b,h,((20,-8),(0,0),(20,8)),
                          fuellfarbe='white'))
        # Raute ?
        if not(self.GibTyp() in (1,2)): return figur
        if self.GibTyp()==1: ff='white'
        if self.GibTyp()==2: ff='black'
        if self.__anschluss[0]=='O':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(15,-5),(30,0),(15,5)),fuellfarbe=ff))
        elif self.__anschluss[0]=='W':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-15,-5),(-30,0),(-15,5)),fuellfarbe=ff))
        elif self.__anschluss[0]=='S':
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-5,15),(0,30),(5,15)),fuellfarbe=ff))
        else:                   # 'N'
            figur.FuegeHinzu(
                N_Eck(0,0,((0,0),(-5,-15),(0,-30),(5,-15)),fuellfarbe=ff))
        return figur
            
            
### ----- kd_verbinder -------------------------------------
class KD_Verbinder(object):
    """entkoppelt vom grafischen Aspekt die Eigenschaften und Methoden
    fuer Verbinder von Klassen in Klassendiagrammen"""

    def __init__(self,
                 startKarte,
                 startKante,
                 zielKarte,
                 zielKante,
                 name=''):
        self.__name = name
        self.__startKarte=startKarte
        self.__startKante=startKante
        self.__startTeiler=50 # Prozent der Kante
        self.__zielKarte=zielKarte
        self.__zielKante=zielKante
        self.__zielTeiler=50 # Prozent der Kante
        self.__verbinder=None # der grafische Aspekt
        self.__verbindertyp=0
        self.__richtung=0 # fuer Pfeil zum Ziel
        self.__typ=0 #

    ## Get-Methoden
    def GibDaten(self): return self.GibStartWerte()+self.GibZielWerte()
    def GibNamen(self): return self.__name
    def GibStartKarte(self): return self.__startKarte
    def GibStartKante(self): return self.__startKante
    def GibStartTeiler(self): return self.__startTeiler
    def GibZielKarte(self): return self.__zielKarte
    def GibZielKante(self): return self.__zielKante
    def GibZielTeiler(self): return self.__zielTeiler
    def GibVerbinder(self): return self.__verbinder
    def GibDatenTextlich(self):
        '''gibt die Daten als Strings zurueck'''
        dat=self.GibStartKarte().GibNamen(),self.__startKante,str(self.__startTeiler),\
            self.GibZielKarte().GibNamen(),self.GibZielKante(),str(self.__zielTeiler),\
            self.GibNamen(),str(self.__typ),str(self.__verbindertyp),str(self.__richtung)
        return dat
    def GibTypNamen(self):
        if self.__typ==0: return 'Assoziation'
        if self.__typ==1: return 'Aggregation'
        if self.__typ==2: return 'Komposition'
        if self.__typ==3: return 'Vererbung'
        return None
   
    def SetzeVerbinderTyp(self, verbinderTyp=0):
        '''setzt den Typ des grafischen Aspekts
        0: Linie,  1: einfacher Winkel,  2: Doppelwinkel'''
        if self.__verbinder!=None:
            self.__verbinder.Verberge()
        self.__verbindertyp=verbinderTyp
        if self.__startKante==None: return
        if self.__zielKante==None: return
        x0,y0,x1,y1 = self.GibDaten()
        von_nach=self.__startKante+self.__zielKante
        if verbinderTyp==0:
            self.__verbinder=Verbinder(x0,y0,x1,y1,self.__name,
                                       self.__richtung,self.__typ)
        elif verbinderTyp==1:
            self.__verbinder=Verbinder1(x0,y0,x1,y1,self.__name,von_nach,
                                        self.__richtung,self.__typ)
        else:
            self.__verbinder=Verbinder2(x0,y0,x1,y1,self.__name,von_nach,
                                        self.__richtung,self.__typ)
        self.__verbinder.Zeige()

    def SetzeRichtung(self, richtung):
        '''ermoeglicht das Setzen einer Richtung, also einer Pfeilspitze
        Vererbung hat keine zusätzliche Richtung'''
        if self.__verbinder==None: return
        if self.__typ==3: return
        self.__richtung = richtung
        self.GibVerbinder().SetzeRichtung(richtung)
    
    def SetzeTyp(self, typ):
        '''ermoeglicht das Setzen von
        0: einfache Assoziation,
        1: Aggregation,
        2: Komposition oder
        3: Vererbung'''
        if self.__verbinder==None: return
        self.__verbinder.Verberge()
        self.__typ=typ
        self.__verbinder.SetzeTyp(typ)
        self.__verbinder.Zeige()

    def SetzeStartKarte(self, klassenkarte, kante=None):
        self.__startKarte=klassenkarte
        if kante!=None: self.__startKante=kante
        
    def SetzeStartKante(self, kante='N'):
        '''eine andere Kante verwenden'''
        if self.__startKarte==None: return
        self.__startKante=kante
        self.SetzeVerbinderTyp(self.__verbindertyp)
        self.UpdateVerbinder()
        
    def GibStartWerte(self):
        if self.__startKarte==None: return None
        if self.__startKante==None: return None
        x,y,b,h=self.__startKarte.GibAbmessungen()
        ##print('Start',x,y,b,h)
        if self.__startKante=='N':
            return x+b*self.__startTeiler/100, y
        elif self.__startKante=='S':
            return x+b*self.__startTeiler/100, y+h
        elif self.__startKante=='W':
            return x, y+h*self.__startTeiler/100
        else:                # 'O'
            return x+b, y+h*self.__startTeiler/100

    def SetzeZielKarte(self, klassenkarte, kante=None):
        self.__zielKarte=klassenkarte
        if kante!=None: self.__zielKante=kante
        
    def GibZielWerte(self):
        if self.__zielKarte==None: return None
        if self.__zielKante==None: return None
        x,y,b,h=self.__zielKarte.GibAbmessungen()
        ##print('Ziel',x,y,b,h)
        if self.__zielKante=='N':
            return x+b*self.__zielTeiler/100, y
        elif self.__zielKante=='S':
            return x+b*self.__zielTeiler/100, y+h
        elif self.__zielKante=='W':
            return x, y+h*self.__zielTeiler/100
        else:                # 'O'
            return x+b, y+h*self.__zielTeiler/100

    def SetzeZielKante(self, kante='N'):
        '''eine andere Kante verwenden'''
        if self.__zielKarte==None: return
        self.__zielKante=kante
        self.SetzeVerbinderTyp(self.__verbindertyp)
        self.UpdateVerbinder()
        
    def NeuerTeilerStartKante(self, teiler=50):
        '''setzt den prozentual Anteil auf der Kante neu'''
        self.__startTeiler=teiler
        self.UpdateVerbinder()

    def NeuerTeilerZielKante(self, teiler=50):
        '''setzt den prozentual Anteil auf der Kante neu'''
        self.__zielTeiler=teiler
        self.UpdateVerbinder()
        
    def NeuerName(self, name):
        '''fuegt dem Pfeil einen Namen hinzu'''
        pass ## T O D O

    def UpdateVerbinder(self):
        self.__verbinder.Verberge()
        self.__verbinder.SetzeStart(self.GibStartWerte())
        self.__verbinder.SetzeZiel(self.GibZielWerte())
        self.__verbinder.Zeige()

    def Verberge(self):
        self.__verbinder.Verberge()

    def Zeige(self):
        self.__verbinder.Zeige()


### ----- istEin -------------------------------------------
class VererbungsPfeil(KD_Verbinder):
    """Vererbungspfeil fuer UML-Diagramme"""

    def __init__(self,
                 startKarte,
                 startKante,
                 zielKarte,
                 zielKante,
                 name='',
                 typ=3, # steht immer fuer Vererbung
                 verbindertyp=0
                 ):
        KD_Verbinder.__init__(self,
                              startKarte,
                              startKante,
                              zielKarte,
                              zielKante,
                              name)
        self.SetzeVerbinderTyp(verbindertyp)
        self.SetzeRichtung(0) # kein Richtungspfeil bei Vererbung
        self.SetzeTyp(3) ## steht fuer Vererbung
        

### ----- Assoziation --------------------------------------
class Assoziation(KD_Verbinder):
    """Assoziation fuer UML-Diagramme"""

    def __init__(self,
                 startKarte,
                 startKante,
                 zielKarte,
                 zielKante,
                 name='',
                 typ=0,
                 verbindertyp=0,
                 richtung=0   # alternativ 1
                 ):
        KD_Verbinder.__init__(self,
                              startKarte,
                              startKante,
                              zielKarte,
                              zielKante,
                              name)
        self.SetzeVerbinderTyp(verbindertyp)
        self.SetzeTyp(typ)
        # 0: 'einfach', 1: 'Aggregation', 2: 'Komposition':
        self.SetzeRichtung(richtung)
